@@ -2,24 +2,28 @@
  * ra.c :  routines for interacting with the RA layer
  *
  * ====================================================================
- * Copyright (c) 2000-2007 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
 
 
 #include <apr_pools.h>
-#include <assert.h>
 
 #include "svn_error.h"
 #include "svn_pools.h"
@@ -27,6 +31,7 @@
 #include "svn_sorts.h"
 #include "svn_ra.h"
 #include "svn_client.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_props.h"
 #include "svn_mergeinfo.h"
@@ -36,43 +41,42 @@
 #include "svn_private_config.h"
 #include "private/svn_wc_private.h"
 
+
+/* This is the baton that we pass svn_ra_open3(), and is associated with
+   the callback table we provide to RA. */
+typedef struct callback_baton_t
+{
+  /* Holds the directory that corresponds to the REPOS_URL at svn_ra_open3()
+     time. When callbacks specify a relative path, they are joined with
+     this base directory. */
+  const char *base_dir_abspath;
+
+  /* When true, makes sure temporary files are created
+     outside the working copy. */
+  svn_boolean_t read_only_wc;
+
+  /* An array of svn_client_commit_item3_t * structures, present only
+     during working copy commits. */
+  const apr_array_header_t *commit_items;
+
+  /* A client context. */
+  svn_client_ctx_t *ctx;
+
+  /* The pool to use for session-related items. */
+  apr_pool_t *pool;
+
+} callback_baton_t;
+
 
 
-static svn_error_t *
-open_admin_tmp_file(apr_file_t **fp,
-                    void *callback_baton,
-                    apr_pool_t *pool)
-{
-  svn_client__callback_baton_t *cb = callback_baton;
-
-  SVN_ERR(svn_wc_create_tmp_file2(fp, NULL, cb->base_dir,
-                                  svn_io_file_del_on_close, pool));
-
-  return SVN_NO_ERROR;
-}
-
-
 static svn_error_t *
 open_tmp_file(apr_file_t **fp,
               void *callback_baton,
               apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *cb = callback_baton;
-  const char *truepath;
-
-  if (cb->base_dir && ! cb->read_only_wc)
-    truepath = apr_pstrdup(pool, cb->base_dir);
-  else
-    SVN_ERR(svn_io_temp_dir(&truepath, pool));
-
-  /* Tack on a made-up filename. */
-  truepath = svn_path_join(truepath, "tempfile", pool);
-
-  /* Open a unique file;  use APR_DELONCLOSE. */
-  SVN_ERR(svn_io_open_unique_file2(fp, NULL, truepath, ".tmp",
-                                   svn_io_file_del_on_close, pool));
-
-  return SVN_NO_ERROR;
+  return svn_error_trace(svn_io_open_unique_file3(fp, NULL, NULL,
+                                  svn_io_file_del_on_pool_cleanup,
+                                  pool, pool));
 }
 
 
@@ -84,7 +88,9 @@ get_wc_prop(void *baton,
             const svn_string_t **value,
             apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *cb = baton;
+  callback_baton_t *cb = baton;
+  const char *local_abspath = NULL;
+  svn_error_t *err;
 
   *value = NULL;
 
@@ -96,24 +102,39 @@ get_wc_prop(void *baton,
       for (i = 0; i < cb->commit_items->nelts; i++)
         {
           svn_client_commit_item3_t *item
-            = APR_ARRAY_IDX(cb->commit_items, i,
-                            svn_client_commit_item3_t *);
-          if (! strcmp(relpath,
-                       svn_path_uri_decode(item->url, pool)))
-            return svn_wc_prop_get(value, name, item->path, cb->base_access,
-                                   pool);
+            = APR_ARRAY_IDX(cb->commit_items, i, svn_client_commit_item3_t *);
+
+          if (! strcmp(relpath, item->session_relpath))
+            {
+              SVN_ERR_ASSERT(svn_dirent_is_absolute(item->path));
+              local_abspath = item->path;
+              break;
+            }
         }
 
-      return SVN_NO_ERROR;
+      /* Commits can only query relpaths in the commit_items list
+         since the commit driver traverses paths as they are, or will
+         be, in the repository.  Non-commits query relpaths in the
+         working copy. */
+      if (! local_abspath)
+        return SVN_NO_ERROR;
     }
 
   /* If we don't have a base directory, then there are no properties. */
-  else if (cb->base_dir == NULL)
+  else if (cb->base_dir_abspath == NULL)
     return SVN_NO_ERROR;
 
-  return svn_wc_prop_get(value, name,
-                         svn_path_join(cb->base_dir, relpath, pool),
-                         cb->base_access, pool);
+  else
+    local_abspath = svn_dirent_join(cb->base_dir_abspath, relpath, pool);
+
+  err = svn_wc_prop_get2(value, cb->ctx->wc_ctx, local_abspath, name,
+                         pool, pool);
+  if (err && err->apr_err == SVN_ERR_WC_PATH_NOT_FOUND)
+    {
+      svn_error_clear(err);
+      err = NULL;
+    }
+  return svn_error_trace(err);
 }
 
 /* This implements the 'svn_ra_push_wc_prop_func_t' interface. */
@@ -124,7 +145,7 @@ push_wc_prop(void *baton,
              const svn_string_t *value,
              apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *cb = baton;
+  callback_baton_t *cb = baton;
   int i;
 
   /* If we're committing, search through the commit_items list for a
@@ -133,24 +154,21 @@ push_wc_prop(void *baton,
     return svn_error_createf
       (SVN_ERR_UNSUPPORTED_FEATURE, NULL,
        _("Attempt to set wc property '%s' on '%s' in a non-commit operation"),
-       name, svn_path_local_style(relpath, pool));
+       name, svn_dirent_local_style(relpath, pool));
 
   for (i = 0; i < cb->commit_items->nelts; i++)
     {
       svn_client_commit_item3_t *item
         = APR_ARRAY_IDX(cb->commit_items, i, svn_client_commit_item3_t *);
 
-      if (strcmp(relpath, svn_path_uri_decode(item->url, pool)) == 0)
+      if (strcmp(relpath, item->session_relpath) == 0)
         {
-          apr_pool_t *cpool = item->incoming_prop_changes->pool;
-          svn_prop_t *prop = apr_palloc(cpool, sizeof(*prop));
+          apr_pool_t *changes_pool = item->incoming_prop_changes->pool;
+          svn_prop_t *prop = apr_palloc(changes_pool, sizeof(*prop));
 
-          prop->name = apr_pstrdup(cpool, name);
+          prop->name = apr_pstrdup(changes_pool, name);
           if (value)
-            {
-              prop->value
-                = svn_string_ncreate(value->data, value->len, cpool);
-            }
+            prop->value = svn_string_dup(value, changes_pool);
           else
             prop->value = NULL;
 
@@ -173,19 +191,10 @@ set_wc_prop(void *baton,
             const svn_string_t *value,
             apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *cb = baton;
-  svn_wc_adm_access_t *adm_access;
-  const svn_wc_entry_t *entry;
-  const char *full_path = svn_path_join(cb->base_dir, path, pool);
+  callback_baton_t *cb = baton;
+  const char *local_abspath;
 
-  SVN_ERR(svn_wc__entry_versioned(&entry, full_path, cb->base_access, FALSE,
-                                 pool));
-
-  SVN_ERR(svn_wc_adm_retrieve(&adm_access, cb->base_access,
-                              (entry->kind == svn_node_dir
-                               ? full_path
-                               : svn_path_dirname(full_path, pool)),
-                              pool));
+  local_abspath = svn_dirent_join(cb->base_dir_abspath, path, pool);
 
   /* We pass 1 for the 'force' parameter here.  Since the property is
      coming from the repository, we definitely want to accept it.
@@ -195,40 +204,14 @@ set_wc_prop(void *baton,
      right, but the conflict would remind the user to make sure.
      Unfortunately, we don't have a clean mechanism for doing that
      here, so we just set the property and hope for the best. */
-  return svn_wc_prop_set2(name, value, full_path, adm_access, TRUE, pool);
-}
-
-
-struct invalidate_wcprop_walk_baton
-{
-  /* The wcprop to invalidate. */
-  const char *prop_name;
-
-  /* Access baton for the top of the walk. */
-  svn_wc_adm_access_t *base_access;
-};
-
-
-/* This implements the `found_entry' prototype in
-   `svn_wc_entry_callbacks_t'. */
-static svn_error_t *
-invalidate_wcprop_for_entry(const char *path,
-                            const svn_wc_entry_t *entry,
-                            void *walk_baton,
-                            apr_pool_t *pool)
-{
-  struct invalidate_wcprop_walk_baton *wb = walk_baton;
-  svn_wc_adm_access_t *entry_access;
-
-  SVN_ERR(svn_wc_adm_retrieve(&entry_access, wb->base_access,
-                              ((entry->kind == svn_node_dir)
-                               ? path
-                               : svn_path_dirname(path, pool)),
-                              pool));
-  /* It doesn't matter if we pass 0 or 1 for force here, since
-     property deletion is always permitted. */
-  return svn_wc_prop_set2(wb->prop_name, NULL, path, entry_access,
-                          FALSE, pool);
+  return svn_error_trace(svn_wc_prop_set4(cb->ctx->wc_ctx, local_abspath,
+                                          name,
+                                          value, svn_depth_empty,
+                                          TRUE /* skip_checks */,
+                                          NULL /* changelist_filter */,
+                                          NULL, NULL /* cancellation */,
+                                          NULL, NULL /* notification */,
+                                          pool));
 }
 
 
@@ -239,32 +222,29 @@ invalidate_wc_props(void *baton,
                     const char *prop_name,
                     apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *cb = baton;
-  svn_wc_entry_callbacks2_t walk_callbacks = { invalidate_wcprop_for_entry,
-                              svn_client__default_walker_error_handler };
-  struct invalidate_wcprop_walk_baton wb;
-  svn_wc_adm_access_t *adm_access;
+  callback_baton_t *cb = baton;
+  const char *local_abspath;
 
-  wb.base_access = cb->base_access;
-  wb.prop_name = prop_name;
+  local_abspath = svn_dirent_join(cb->base_dir_abspath, path, pool);
 
-  path = svn_path_join(cb->base_dir, path, pool);
-  SVN_ERR(svn_wc_adm_probe_retrieve(&adm_access, cb->base_access, path,
-                                    pool));
-  SVN_ERR(svn_wc_walk_entries3(path, adm_access, &walk_callbacks, &wb,
-                               svn_depth_infinity, FALSE,
-                               cb->ctx->cancel_func, cb->ctx->cancel_baton,
-                               pool));
-
-  return SVN_NO_ERROR;
+  /* It's easier just to clear the whole dav_cache than to remove
+     individual items from it recursively like this.  And since we
+     know that the RA providers that ship with Subversion only
+     invalidate the one property they use the most from this cache,
+     and that we're intentionally trying to get away from the use of
+     the cache altogether anyway, there's little to lose in wiping the
+     whole cache.  Is it the most well-behaved approach to take?  Not
+     so much.  We choose not to care.  */
+  return svn_error_trace(svn_wc__node_clear_dav_cache_recursive(
+                              cb->ctx->wc_ctx, local_abspath, pool));
 }
 
 
 static svn_error_t *
 cancel_callback(void *baton)
 {
-  svn_client__callback_baton_t *b = baton;
-  return (b->ctx->cancel_func)(b->ctx->cancel_baton);
+  callback_baton_t *b = baton;
+  return svn_error_trace((b->ctx->cancel_func)(b->ctx->cancel_baton));
 }
 
 
@@ -273,27 +253,34 @@ get_client_string(void *baton,
                   const char **name,
                   apr_pool_t *pool)
 {
-  svn_client__callback_baton_t *b = baton;
+  callback_baton_t *b = baton;
   *name = apr_pstrdup(pool, b->ctx->client_name);
   return SVN_NO_ERROR;
 }
 
+
+#define SVN_CLIENT__MAX_REDIRECT_ATTEMPTS 3 /* ### TODO:  Make configurable. */
+
 svn_error_t *
 svn_client__open_ra_session_internal(svn_ra_session_t **ra_session,
+                                     const char **corrected_url,
                                      const char *base_url,
-                                     const char *base_dir,
-                                     svn_wc_adm_access_t *base_access,
-                                     apr_array_header_t *commit_items,
+                                     const char *base_dir_abspath,
+                                     const apr_array_header_t *commit_items,
                                      svn_boolean_t use_admin,
                                      svn_boolean_t read_only_wc,
                                      svn_client_ctx_t *ctx,
                                      apr_pool_t *pool)
 {
   svn_ra_callbacks2_t *cbtable = apr_pcalloc(pool, sizeof(*cbtable));
-  svn_client__callback_baton_t *cb = apr_pcalloc(pool, sizeof(*cb));
+  callback_baton_t *cb = apr_pcalloc(pool, sizeof(*cb));
   const char *uuid = NULL;
 
-  cbtable->open_tmp_file = use_admin ? open_admin_tmp_file : open_tmp_file;
+  SVN_ERR_ASSERT(base_dir_abspath != NULL || ! use_admin);
+  SVN_ERR_ASSERT(base_dir_abspath == NULL
+                        || svn_dirent_is_absolute(base_dir_abspath));
+
+  cbtable->open_tmp_file = open_tmp_file;
   cbtable->get_wc_prop = use_admin ? get_wc_prop : NULL;
   cbtable->set_wc_prop = read_only_wc ? NULL : set_wc_prop;
   cbtable->push_wc_prop = commit_items ? push_wc_prop : NULL;
@@ -304,28 +291,86 @@ svn_client__open_ra_session_internal(svn_ra_session_t **ra_session,
   cbtable->cancel_func = ctx->cancel_func ? cancel_callback : NULL;
   cbtable->get_client_string = get_client_string;
 
-  cb->base_dir = base_dir;
-  cb->base_access = base_access;
+  cb->base_dir_abspath = base_dir_abspath;
   cb->read_only_wc = read_only_wc;
   cb->pool = pool;
   cb->commit_items = commit_items;
   cb->ctx = ctx;
 
-  if (base_access)
+  if (base_dir_abspath)
     {
-      const svn_wc_entry_t *entry;
-      
-      SVN_ERR(svn_wc_entry(&entry, base_dir, base_access, FALSE, pool));
+      svn_error_t *err = svn_wc__node_get_repos_info(NULL, &uuid, ctx->wc_ctx,
+                                                     base_dir_abspath,
+                                                     pool, pool);
 
-      if (entry && entry->uuid)
-        uuid = entry->uuid;
+      if (err && (err->apr_err == SVN_ERR_WC_NOT_WORKING_COPY
+                  || err->apr_err == SVN_ERR_WC_PATH_NOT_FOUND
+                  || err->apr_err == SVN_ERR_WC_UPGRADE_REQUIRED))
+        {
+          svn_error_clear(err);
+          uuid = NULL;
+        }
+      else
+        SVN_ERR(err);
     }
 
-  SVN_ERR(svn_ra_open3(ra_session, base_url, uuid, cbtable, cb,
-                       ctx->config, pool));
+  /* If the caller allows for auto-following redirections, and the
+     RA->open() call above reveals a CORRECTED_URL, try the new URL.
+     We'll do this in a loop up to some maximum number follow-and-retry
+     attempts.  */
+  if (corrected_url)
+    {
+      apr_hash_t *attempted = apr_hash_make(pool);
+      int attempts_left = SVN_CLIENT__MAX_REDIRECT_ATTEMPTS;
+
+      *corrected_url = NULL;
+      while (attempts_left--)
+        {
+          const char *corrected = NULL;
+
+          /* Try to open the RA session.  If this is our last attempt,
+             don't accept corrected URLs from the RA provider. */
+          SVN_ERR(svn_ra_open4(ra_session,
+                               attempts_left == 0 ? NULL : &corrected,
+                               base_url, uuid, cbtable, cb, ctx->config, pool));
+
+          /* No error and no corrected URL?  We're done here. */
+          if (! corrected)
+            break;
+
+          /* Notify the user that a redirect is being followed. */
+          if (ctx->notify_func2 != NULL)
+            {
+              svn_wc_notify_t *notify =
+                svn_wc_create_notify_url(corrected,
+                                         svn_wc_notify_url_redirect, pool);
+              (*ctx->notify_func2)(ctx->notify_baton2, notify, pool);
+            }
+
+          /* Our caller will want to know what our final corrected URL was. */
+          *corrected_url = corrected;
+
+          /* Make sure we've not attempted this URL before. */
+          if (apr_hash_get(attempted, corrected, APR_HASH_KEY_STRING))
+            return svn_error_createf(SVN_ERR_CLIENT_CYCLE_DETECTED, NULL,
+                                     _("Redirect cycle detected for URL '%s'"),
+                                     corrected);
+
+          /* Remember this CORRECTED_URL so we don't wind up in a loop. */
+          apr_hash_set(attempted, corrected, APR_HASH_KEY_STRING, (void *)1);
+          base_url = corrected;
+        }
+    }
+  else
+    {
+      SVN_ERR(svn_ra_open4(ra_session, NULL, base_url,
+                           uuid, cbtable, cb, ctx->config, pool));
+    }
 
   return SVN_NO_ERROR;
-}
+ }
+#undef SVN_CLIENT__MAX_REDIRECT_ATTEMPTS
+
 
 svn_error_t *
 svn_client_open_ra_session(svn_ra_session_t **session,
@@ -333,8 +378,10 @@ svn_client_open_ra_session(svn_ra_session_t **session,
                            svn_client_ctx_t *ctx,
                            apr_pool_t *pool)
 {
-  return svn_client__open_ra_session_internal(session, url, NULL, NULL, NULL,
-                                              FALSE, TRUE, ctx, pool);
+  return svn_error_trace(
+             svn_client__open_ra_session_internal(session, NULL, url,
+                                                  NULL, NULL, FALSE, TRUE,
+                                                  ctx, pool));
 }
 
 
@@ -348,9 +395,9 @@ svn_client_uuid_from_url(const char **uuid,
   apr_pool_t *subpool = svn_pool_create(pool);
 
   /* use subpool to create a temporary RA session */
-  SVN_ERR(svn_client__open_ra_session_internal(&ra_session, url,
+  SVN_ERR(svn_client__open_ra_session_internal(&ra_session, NULL, url,
                                                NULL, /* no base dir */
-                                               NULL, NULL, FALSE, TRUE,
+                                               NULL, FALSE, TRUE,
                                                ctx, subpool));
 
   SVN_ERR(svn_ra_get_uuid2(ra_session, uuid, pool));
@@ -363,99 +410,53 @@ svn_client_uuid_from_url(const char **uuid,
 
 
 svn_error_t *
-svn_client_uuid_from_path(const char **uuid,
-                          const char *path,
-                          svn_wc_adm_access_t *adm_access,
-                          svn_client_ctx_t *ctx,
-                          apr_pool_t *pool)
+svn_client_uuid_from_path2(const char **uuid,
+                           const char *local_abspath,
+                           svn_client_ctx_t *ctx,
+                           apr_pool_t *result_pool,
+                           apr_pool_t *scratch_pool)
 {
-  const svn_wc_entry_t *entry;
-  svn_boolean_t is_root;
-
-  SVN_ERR(svn_wc__entry_versioned(&entry, path, adm_access,
-                                 TRUE,  /* show deleted */ pool));
-
-  if (entry->uuid)
-    {
-      *uuid = entry->uuid;
-      return SVN_NO_ERROR;
-    }
-
-  /* ## Probably never reached after the 1.6/1.7 WC rewrite */
-  
-  SVN_ERR(svn_wc_is_wc_root(&is_root, path, adm_access, pool));
-
-  if (!is_root)
-    {
-      /* Workingcopies have a single uuid, as all contents is from a single
-         repository */
-
-      svn_error_t *err;
-      svn_wc_adm_access_t *parent_access;
-      const char *parent = svn_path_dirname(path, pool);
-
-      /* Open the parents administrative area to fetch the uuid.
-         Subversion 1.0 and later have the uuid in every checkout root */
-
-      SVN_ERR(svn_wc_adm_open3(&parent_access, NULL, parent, FALSE, 0,
-                               ctx->cancel_func, ctx->cancel_baton, pool));
-      
-      err = svn_client_uuid_from_path(uuid, svn_path_dirname(path, pool),
-                                      parent_access, ctx, pool);
-
-      svn_error_clear(svn_wc_adm_close(parent_access));
-
-      return err;
-    }
-    
-  /* We may have a workingcopy without uuid */     
-  if (entry->url)
-    {
-      /* You can enter this case by copying a new subdirectory with 1.0-1.5
-       * # svn mkdir newdir
-       * # cp newdir /tmp/new-wc
-       * and then check /tmp/new-wc
-       *
-       * See also:
-       * http://subversion.tigris.org/servlets/ReadMsg?list=dev&msgNo=101831
-       * Message-ID: <877jgjtkus.fsf@debian2.lan> */
-
-      /* fallback to using the network. */
-      SVN_ERR(svn_client_uuid_from_url(uuid, entry->url, ctx, pool));
-    }
-  else
-    {
-      return svn_error_createf(SVN_ERR_ENTRY_MISSING_URL, NULL,
-                               _("'%s' has no URL"),
-                               svn_path_local_style(path, pool));
-    }
-
-  return SVN_NO_ERROR;
+  return svn_error_trace(
+    svn_wc__node_get_repos_info(NULL, uuid, ctx->wc_ctx, local_abspath,
+                                result_pool, scratch_pool));
 }
 
+
 
 
+/* Convert a path or URL for display: if it is a local path, convert it to
+ * the local path style; if it is a URL, return it unchanged. */
+static const char *
+path_or_url_local_style(const char *path_or_url,
+                        apr_pool_t *pool)
+{
+  if (svn_path_is_url(path_or_url))
+    return path_or_url;
+  return svn_dirent_local_style(path_or_url, pool);
+}
+
 svn_error_t *
 svn_client__ra_session_from_path(svn_ra_session_t **ra_session_p,
                                  svn_revnum_t *rev_p,
                                  const char **url_p,
                                  const char *path_or_url,
-                                 svn_wc_adm_access_t *base_access,
+                                 const char *base_dir_abspath,
                                  const svn_opt_revision_t *peg_revision_p,
                                  const svn_opt_revision_t *revision,
                                  svn_client_ctx_t *ctx,
                                  apr_pool_t *pool)
 {
   svn_ra_session_t *ra_session;
-  const char *initial_url, *url, *base_dir = NULL;
+  const char *initial_url, *url;
   svn_opt_revision_t *good_rev;
   svn_opt_revision_t peg_revision, start_rev;
   svn_opt_revision_t dead_end_rev;
-  svn_opt_revision_t *ignored_rev, *new_rev;
+  svn_opt_revision_t *ignored_rev;
   svn_revnum_t rev;
-  const char *ignored_url;
+  const char *ignored_url, *corrected_url;
 
-  SVN_ERR(svn_client_url_from_path(&initial_url, path_or_url, pool));
+  SVN_ERR(svn_client_url_from_path2(&initial_url, path_or_url, ctx, pool,
+                                    pool));
   if (! initial_url)
     return svn_error_createf(SVN_ERR_ENTRY_MISSING_URL, NULL,
                              _("'%s' has no URL"), path_or_url);
@@ -467,26 +468,28 @@ svn_client__ra_session_from_path(svn_ra_session_t **ra_session_p,
                                     TRUE,
                                     pool));
 
-  if (base_access)
-    base_dir = svn_wc_adm_access_path(base_access);
-
-  SVN_ERR(svn_client__open_ra_session_internal(&ra_session, initial_url,
-                                               base_dir, base_access, NULL,
-                                               base_access ? TRUE : FALSE,
+  SVN_ERR(svn_client__open_ra_session_internal(&ra_session, &corrected_url,
+                                               initial_url,
+                                               base_dir_abspath, NULL,
+                                               base_dir_abspath != NULL,
                                                FALSE, ctx, pool));
 
+  /* If we got a CORRECTED_URL, we'll want to refer to that as the
+     URL-ized form of PATH_OR_URL from now on. */
+  if (corrected_url && svn_path_is_url(path_or_url))
+    path_or_url = corrected_url;
+
   dead_end_rev.kind = svn_opt_revision_unspecified;
 
   /* Run the history function to get the object's (possibly
      different) url in REVISION. */
-  SVN_ERR(svn_client__repos_locations(&url, &new_rev,
+  SVN_ERR(svn_client__repos_locations(&url, &good_rev,
                                       &ignored_url, &ignored_rev,
                                       ra_session,
                                       path_or_url, &peg_revision,
                                       /* search range: */
                                       &start_rev, &dead_end_rev,
                                       ctx, pool));
-  good_rev = (svn_opt_revision_t *)new_rev;
 
   /* Make the session point to the real URL. */
   SVN_ERR(svn_ra_reparent(ra_session, url, pool));
@@ -494,8 +497,8 @@ svn_client__ra_session_from_path(svn_ra_session_t **ra_session_p,
   /* Resolve good_rev into a real revnum. */
   if (good_rev->kind == svn_opt_revision_unspecified)
     good_rev->kind = svn_opt_revision_head;
-  SVN_ERR(svn_client__get_revision_number(&rev, NULL, ra_session,
-                                          good_rev, url, pool));
+  SVN_ERR(svn_client__get_revision_number(&rev, NULL, ctx->wc_ctx, url,
+                                          ra_session, good_rev, pool));
 
   *ra_session_p = ra_session;
   *rev_p = rev;
@@ -505,23 +508,6 @@ svn_client__ra_session_from_path(svn_ra_session_t **ra_session_p,
 }
 
 
-svn_error_t *
-svn_client__path_relative_to_session(const char **rel_path,
-                                     svn_ra_session_t *ra_session,
-                                     const char *url, 
-                                     apr_pool_t *pool)
-{
-  const char *session_url;
-  SVN_ERR(svn_ra_get_session_url(ra_session, &session_url, pool));
-  if (strcmp(session_url, url) == 0)
-    *rel_path = "";
-  else
-    *rel_path = svn_path_uri_decode(svn_path_is_child(session_url, url, pool), 
-                                    pool);
-  return SVN_NO_ERROR;
-}
-
-
 svn_error_t *
 svn_client__ensure_ra_session_url(const char **old_session_url,
                                   svn_ra_session_t *ra_session,
@@ -567,9 +553,9 @@ gls_receiver(svn_location_segment_t *segment,
 static int
 compare_segments(const void *a, const void *b)
 {
-  const svn_location_segment_t *a_seg 
+  const svn_location_segment_t *a_seg
     = *((const svn_location_segment_t * const *) a);
-  const svn_location_segment_t *b_seg 
+  const svn_location_segment_t *b_seg
     = *((const svn_location_segment_t * const *) b);
   if (a_seg->range_start == b_seg->range_start)
     return 0;
@@ -618,6 +604,7 @@ svn_client__repos_locations(const char **start_url,
   const char *url;
   const char *start_path = NULL;
   const char *end_path = NULL;
+  const char *local_abspath_or_url;
   svn_revnum_t peg_revnum = SVN_INVALID_REVNUM;
   svn_revnum_t start_revnum, end_revnum;
   svn_revnum_t youngest_rev = SVN_INVALID_REVNUM;
@@ -637,36 +624,55 @@ svn_client__repos_locations(const char **start_url,
      the copyfrom information. */
   if (! svn_path_is_url(path))
     {
-      svn_wc_adm_access_t *adm_access;
-      const svn_wc_entry_t *entry;
-      SVN_ERR(svn_wc_adm_probe_open3(&adm_access, NULL, path,
-                                     FALSE, 0, ctx->cancel_func,
-                                     ctx->cancel_baton, pool));
-      SVN_ERR(svn_wc_entry(&entry, path, adm_access, FALSE, pool));
-      SVN_ERR(svn_wc_adm_close(adm_access));
-      if (entry->copyfrom_url && revision->kind == svn_opt_revision_working)
+      SVN_ERR(svn_dirent_get_absolute(&local_abspath_or_url, path, subpool));
+
+      if (revision->kind == svn_opt_revision_working)
         {
-          url = entry->copyfrom_url;
-          peg_revnum = entry->copyfrom_rev;
-          if (!entry->url || strcmp(entry->url, entry->copyfrom_url) != 0)
+          const char *repos_root_url;
+          const char *repos_relpath;
+          svn_boolean_t is_copy;
+
+          SVN_ERR(svn_wc__node_get_origin(&is_copy, &peg_revnum, &repos_relpath,
+                                          &repos_root_url, NULL,
+                                          ctx->wc_ctx, local_abspath_or_url,
+                                          FALSE, subpool, subpool));
+
+          if (repos_relpath)
+            url = svn_path_url_add_component2(repos_root_url, repos_relpath,
+                                              pool);
+          else
+            url = NULL;
+
+          if (url && is_copy && ra_session)
             {
-              /* We can't use the caller provided RA session in this case */
-              ra_session = NULL;
+              const char *session_url;
+              SVN_ERR(svn_ra_get_session_url(ra_session, &session_url,
+                                             subpool));
+
+              if (strcmp(session_url, url) != 0)
+                {
+                  /* We can't use the caller provided RA session now :( */
+                  ra_session = NULL;
+                }
             }
         }
-      else if (entry->url)
-        {
-          url = entry->url;
-        }
       else
+        url = NULL;
+
+      if (! url)
+        SVN_ERR(svn_wc__node_get_url(&url, ctx->wc_ctx,
+                                     local_abspath_or_url, pool, subpool));
+
+      if (!url)
         {
           return svn_error_createf(SVN_ERR_ENTRY_MISSING_URL, NULL,
                                    _("'%s' has no URL"),
-                                   svn_path_local_style(path, pool));
+                                   svn_dirent_local_style(path, pool));
         }
     }
   else
     {
+      local_abspath_or_url = path;
       url = path;
     }
 
@@ -676,23 +682,25 @@ svn_client__repos_locations(const char **start_url,
 
   /* Open a RA session to this URL if we don't have one already. */
   if (! ra_session)
-    SVN_ERR(svn_client__open_ra_session_internal(&ra_session, url, NULL,
-                                                 NULL, NULL, FALSE, TRUE,
+    SVN_ERR(svn_client__open_ra_session_internal(&ra_session, NULL, url, NULL,
+                                                 NULL, FALSE, TRUE,
                                                  ctx, subpool));
 
   /* Resolve the opt_revision_ts. */
   if (peg_revnum == SVN_INVALID_REVNUM)
     SVN_ERR(svn_client__get_revision_number(&peg_revnum, &youngest_rev,
-                                            ra_session, revision, path,
-                                            pool));
+                                            ctx->wc_ctx, local_abspath_or_url,
+                                            ra_session, revision, pool));
 
   SVN_ERR(svn_client__get_revision_number(&start_revnum, &youngest_rev,
-                                          ra_session, start, path, pool));
+                                          ctx->wc_ctx, local_abspath_or_url,
+                                          ra_session, start, pool));
   if (end->kind == svn_opt_revision_unspecified)
     end_revnum = start_revnum;
   else
     SVN_ERR(svn_client__get_revision_number(&end_revnum, &youngest_rev,
-                                            ra_session, end, path, pool));
+                                            ctx->wc_ctx, local_abspath_or_url,
+                                            ra_session, end, pool));
 
   /* Set the output revision variables. */
   *start_revision = apr_pcalloc(pool, sizeof(**start_revision));
@@ -731,7 +739,7 @@ svn_client__repos_locations(const char **start_url,
     return svn_error_createf
       (SVN_ERR_CLIENT_UNRELATED_RESOURCES, NULL,
        _("Unable to find repository location for '%s' in revision %ld"),
-       path, start_revnum);
+       path_or_url_local_style(path, pool), start_revnum);
 
   end_path = apr_hash_get(rev_locs, &end_revnum, sizeof(svn_revnum_t));
   if (! end_path)
@@ -739,22 +747,12 @@ svn_client__repos_locations(const char **start_url,
       (SVN_ERR_CLIENT_UNRELATED_RESOURCES, NULL,
        _("The location for '%s' for revision %ld does not exist in the "
          "repository or refers to an unrelated object"),
-       path, end_revnum);
-
-  /* Repository paths might be absolute, but we want to treat them as
-     relative.
-     ### Aren't they always absolute? */
-  if (start_path[0] == '/')
-    start_path = start_path + 1;
-  if (end_path[0] == '/')
-    end_path = end_path + 1;
+       path_or_url_local_style(path, pool), end_revnum);
 
   /* Set our return variables */
-  *start_url = svn_path_join(repos_url, svn_path_uri_encode(start_path,
-                                                            pool), pool);
+  *start_url = svn_path_url_add_component2(repos_url, start_path + 1, pool);
   if (end->kind != svn_opt_revision_unspecified)
-    *end_url = svn_path_join(repos_url, svn_path_uri_encode(end_path,
-                                                            pool), pool);
+    *end_url = svn_path_url_add_component2(repos_url, end_path + 1, pool);
 
   svn_pool_destroy(subpool);
   return SVN_NO_ERROR;
@@ -776,6 +774,8 @@ svn_client__get_youngest_common_ancestor(const char **ancestor_path,
   svn_revnum_t yc_revision = SVN_INVALID_REVNUM;
   const char *yc_path = NULL;
   svn_opt_revision_t revision1, revision2;
+  svn_boolean_t has_rev_zero_history1;
+  svn_boolean_t has_rev_zero_history2;
 
   revision1.kind = revision2.kind = svn_opt_revision_number;
   revision1.value.number = rev1;
@@ -783,33 +783,32 @@ svn_client__get_youngest_common_ancestor(const char **ancestor_path,
 
   /* We're going to cheat and use history-as-mergeinfo because it
      saves us a bunch of annoying custom data comparisons and such. */
-  SVN_ERR(svn_client__get_history_as_mergeinfo(&history1, path_or_url1,
-                                               &revision1, 
-                                               SVN_INVALID_REVNUM, 
+  SVN_ERR(svn_client__get_history_as_mergeinfo(&history1,
+                                               &has_rev_zero_history1,
+                                               path_or_url1,
+                                               &revision1,
                                                SVN_INVALID_REVNUM,
-                                               NULL, NULL, ctx, pool));
-  SVN_ERR(svn_client__get_history_as_mergeinfo(&history2, path_or_url2,
-                                               &revision2, 
-                                               SVN_INVALID_REVNUM, 
                                                SVN_INVALID_REVNUM,
-                                               NULL, NULL, ctx, pool));
+                                               NULL, ctx, pool));
+  SVN_ERR(svn_client__get_history_as_mergeinfo(&history2,
+                                               &has_rev_zero_history2,
+                                               path_or_url2,
+                                               &revision2,
+                                               SVN_INVALID_REVNUM,
+                                               SVN_INVALID_REVNUM,
+                                               NULL, ctx, pool));
 
   /* Loop through the first location's history, check for overlapping
      paths and ranges in the second location's history, and
      remembering the youngest matching location. */
-  for (hi = apr_hash_first(NULL, history1); hi; hi = apr_hash_next(hi))
+  for (hi = apr_hash_first(pool, history1); hi; hi = apr_hash_next(hi))
     {
-      const void *key;
-      apr_ssize_t klen;
-      void *val;
-      const char *path;
-      apr_array_header_t *ranges1, *ranges2, *common;
+      const char *path = svn__apr_hash_index_key(hi);
+      apr_ssize_t path_len = svn__apr_hash_index_klen(hi);
+      apr_array_header_t *ranges1 = svn__apr_hash_index_val(hi);
+      apr_array_header_t *ranges2, *common;
 
-      apr_hash_this(hi, &key, &klen, &val);
-      path = key;
-      ranges1 = val;
-
-      ranges2 = apr_hash_get(history2, key, klen);
+      ranges2 = apr_hash_get(history2, path, path_len);
       if (ranges2)
         {
           /* We have a path match.  Now, did our two histories share
@@ -830,6 +829,14 @@ svn_client__get_youngest_common_ancestor(const char **ancestor_path,
         }
     }
 
+  /* It's possible that PATH_OR_URL1 and PATH_OR_URL2's only common
+     history is revision 0. */
+  if (!yc_path && has_rev_zero_history1 && has_rev_zero_history2)
+    {
+      yc_path = "/";
+      yc_revision = 0;
+    }
+
   *ancestor_path = yc_path;
   *ancestor_revision = yc_revision;
   return SVN_NO_ERROR;
